from __future__ import annotations

from tree_sitter import Language, Query, QueryCursor
import tree_sitter_typescript as tsts

try:
    TS_LANG: Language | None = Language(tsts.language_typescript())
except Exception:
    TS_LANG = None

_SAFE_EXEC_OBJECTS: set[str] = {
    "regex",
    "re",
    "regexp",
    "pattern",
    "reg",
    "db",
    "stmt",
    "query",
    "statement",
    "cursor",
    "conn",
    "connection",
}


def _get_text(source: bytes, node) -> str:
    return source[node.start_byte : node.end_byte].decode("utf-8", errors="replace")


def scan_danger(
    root_node, file_path: str, lang: "Language | None" = None
) -> list[dict]:
    findings: list[dict] = []
    if lang is None:
        lang = TS_LANG
    if not lang:
        return []

    source_bytes: bytes = root_node.text

    def check(pattern: str, cap_name: str, rule: str, sev: str, msg: str) -> None:
        try:
            query = Query(lang, pattern)
            cursor = QueryCursor(query)
            captures = cursor.captures(root_node)
            nodes = captures.get(cap_name, [])

            for node in nodes:
                line = node.start_point[0] + 1
                findings.append(
                    {
                        "rule_id": rule,
                        "severity": sev,
                        "message": msg,
                        "file": str(file_path),
                        "line": line,
                        "col": 0,
                    }
                )
        except Exception:
            pass

    check(
        '(call_expression function: (identifier) @eval (#eq? @eval "eval"))',
        "eval",
        "SKY-D501",
        "CRITICAL",
        "Use of eval() detected",
    )

    check(
        '(assignment_expression left: (member_expression property: (property_identifier) @xss (#eq? @xss "innerHTML")))',
        "xss",
        "SKY-D502",
        "HIGH",
        "Unsafe innerHTML assignment",
    )

    check(
        '(call_expression function: (member_expression property: (property_identifier) @dw (#eq? @dw "write") object: (identifier) @obj (#eq? @obj "document")))',
        "dw",
        "SKY-D503",
        "HIGH",
        "document.write() can lead to XSS vulnerabilities",
    )

    check(
        '(new_expression constructor: (identifier) @fn (#eq? @fn "Function"))',
        "fn",
        "SKY-D504",
        "CRITICAL",
        "new Function() is equivalent to eval()",
    )

    for fn_name in ("setTimeout", "setInterval"):
        check(
            f'(call_expression function: (identifier) @timer (#eq? @timer "{fn_name}") arguments: (arguments (string) @str))',
            "str",
            "SKY-D505",
            "HIGH",
            f"{fn_name}() with string argument is equivalent to eval()",
        )

    _check_exec(root_node, source_bytes, file_path, findings, lang)

    check(
        '(assignment_expression left: (member_expression property: (property_identifier) @oh (#eq? @oh "outerHTML")))',
        "oh",
        "SKY-D507",
        "HIGH",
        "Unsafe outerHTML assignment",
    )

    # SKY-D509: dangerouslySetInnerHTML
    check(
        '(jsx_attribute (property_identifier) @attr (#eq? @attr "dangerouslySetInnerHTML"))',
        "attr",
        "SKY-D509",
        "HIGH",
        "dangerouslySetInnerHTML bypasses React's XSS protections",
    )

    # SKY-D510: Prototype pollution via __proto__
    check(
        '(member_expression property: (property_identifier) @proto (#eq? @proto "__proto__"))',
        "proto",
        "SKY-D510",
        "HIGH",
        "Prototype pollution via __proto__ access",
    )

    _check_hardcoded_secrets(root_node, source_bytes, file_path, findings)
    _check_fetch_ssrf(root_node, source_bytes, file_path, findings, lang)
    _check_weak_crypto(root_node, source_bytes, file_path, findings, lang)
    _check_open_redirect(root_node, source_bytes, file_path, findings, lang)
    _check_sql_template_injection(root_node, source_bytes, file_path, findings, lang)

    return findings


_SECRET_PREFIXES = (
    "sk-",
    "sk_live_",
    "sk_test_",
    "ghp_",
    "gho_",
    "ghu_",
    "ghs_",
    "ghr_",
    "xoxb-",
    "xoxp-",
    "xoxa-",
    "AKIA",
    "eyJ",
)


def _check_hardcoded_secrets(
    root_node, source: bytes, file_path: str, findings: list[dict]
) -> None:
    def _walk(node):
        if node.type in ("string", "template_string"):
            text = _get_text(source, node)
            if text and text[0] in ("'", '"', "`"):
                text = text[1:]
            if text and text[-1] in ("'", '"', "`"):
                text = text[:-1]
            if len(text) >= 16:
                for prefix in _SECRET_PREFIXES:
                    if text.startswith(prefix) or text.lower().startswith(
                        prefix.lower()
                    ):
                        findings.append(
                            {
                                "rule_id": "SKY-D508",
                                "severity": "CRITICAL",
                                "message": "Potential hardcoded secret or API key. Use environment variables instead.",
                                "file": str(file_path),
                                "line": node.start_point[0] + 1,
                                "col": 0,
                            }
                        )
                        return
        for child in node.children:
            _walk(child)

    _walk(root_node)


def _check_fetch_ssrf(
    root_node, source: bytes, file_path: str, findings: list[dict], lang=None
) -> None:
    """SKY-D511: Detect fetch/axios with variable URL (potential SSRF)."""
    if lang is None:
        lang = TS_LANG
    if not lang:
        return

    pattern = '(call_expression function: (identifier) @fn (#eq? @fn "fetch") arguments: (arguments) @args)'
    try:
        query = Query(lang, pattern)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        for node in captures.get("args", []):
            first_arg = None
            for child in node.children:
                if child.type not in ("(", ")", ","):
                    first_arg = child
                    break
            if first_arg and first_arg.type == "identifier":
                findings.append(
                    {
                        "rule_id": "SKY-D511",
                        "severity": "MEDIUM",
                        "message": "fetch() with variable URL — potential SSRF. Validate URL against allowlist.",
                        "file": str(file_path),
                        "line": node.start_point[0] + 1,
                        "col": 0,
                    }
                )
    except Exception:
        pass

    pattern2 = '(call_expression function: (member_expression object: (identifier) @obj (#eq? @obj "axios")) arguments: (arguments) @args)'
    try:
        query = Query(lang, pattern2)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        for node in captures.get("args", []):
            first_arg = None
            for child in node.children:
                if child.type not in ("(", ")", ","):
                    first_arg = child
                    break
            if first_arg and first_arg.type == "identifier":
                findings.append(
                    {
                        "rule_id": "SKY-D511",
                        "severity": "MEDIUM",
                        "message": "axios call with variable URL — potential SSRF. Validate URL against allowlist.",
                        "file": str(file_path),
                        "line": node.start_point[0] + 1,
                        "col": 0,
                    }
                )
    except Exception:
        pass


def _check_weak_crypto(
    root_node, source: bytes, file_path: str, findings: list[dict], lang=None
) -> None:
    """SKY-D513: Detect crypto.createHash('md5') or crypto.createHash('sha1')."""
    if lang is None:
        lang = TS_LANG
    if not lang:
        return
    pattern = '(call_expression function: (member_expression property: (property_identifier) @prop (#eq? @prop "createHash")) arguments: (arguments) @args)'
    try:
        query = Query(lang, pattern)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        for node in captures.get("args", []):
            for child in node.children:
                if child.type == "string":
                    text = _get_text(source, child).strip("'\"")
                    if text in ("md5", "sha1"):
                        findings.append(
                            {
                                "rule_id": "SKY-D513",
                                "severity": "MEDIUM",
                                "message": f"Weak hash algorithm {text.upper()}. Use SHA-256 or better.",
                                "file": str(file_path),
                                "line": node.start_point[0] + 1,
                                "col": 0,
                            }
                        )
                    break
    except Exception:
        pass


def _check_open_redirect(
    root_node, source: bytes, file_path: str, findings: list[dict], lang=None
) -> None:
    """SKY-D515: Detect res.redirect(variable) — Express open redirect."""
    if lang is None:
        lang = TS_LANG
    if not lang:
        return
    pattern = '(call_expression function: (member_expression property: (property_identifier) @prop (#eq? @prop "redirect")) arguments: (arguments) @args)'
    try:
        query = Query(lang, pattern)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        for node in captures.get("args", []):
            first_arg = None
            for child in node.children:
                if child.type not in ("(", ")", ","):
                    first_arg = child
                    break
            if first_arg and first_arg.type not in (
                "string",
                "template_string",
                "number",
            ):
                findings.append(
                    {
                        "rule_id": "SKY-D515",
                        "severity": "HIGH",
                        "message": "Open redirect — res.redirect() with variable argument. Validate redirect target.",
                        "file": str(file_path),
                        "line": node.start_point[0] + 1,
                        "col": 0,
                    }
                )
    except Exception:
        pass


def _check_sql_template_injection(
    root_node, source: bytes, file_path: str, findings: list[dict], lang=None
) -> None:
    """SKY-D516: Detect SQL keywords in template literals passed to query/exec/execute."""
    if lang is None:
        lang = TS_LANG
    if not lang:
        return
    _SQL_KEYWORDS = ("SELECT", "INSERT", "UPDATE", "DELETE", "DROP")
    for method_name in ("query", "exec", "execute"):
        pattern = f'(call_expression function: (member_expression property: (property_identifier) @prop (#eq? @prop "{method_name}")) arguments: (arguments (template_string) @tpl))'
        try:
            query = Query(lang, pattern)
            cursor = QueryCursor(query)
            captures = cursor.captures(root_node)
            for node in captures.get("tpl", []):
                text = _get_text(source, node).upper()
                if any(kw in text for kw in _SQL_KEYWORDS):
                    findings.append(
                        {
                            "rule_id": "SKY-D516",
                            "severity": "CRITICAL",
                            "message": "SQL query built with template literal — risk of SQL injection. Use parameterized queries.",
                            "file": str(file_path),
                            "line": node.start_point[0] + 1,
                            "col": 0,
                        }
                    )
        except Exception:
            pass


def _check_exec(
    root_node, source: bytes, file_path: str, findings: list[dict], lang=None
) -> None:
    if lang is None:
        lang = TS_LANG
    pattern = '(call_expression function: (member_expression object: (identifier) @obj property: (property_identifier) @prop (#eq? @prop "exec")))'
    try:
        query = Query(lang, pattern)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        prop_nodes = captures.get("prop", [])
    except Exception:
        return

    for prop_node in prop_nodes:
        call_node = prop_node.parent
        if call_node is None:
            continue
        obj_node = call_node.child_by_field_name("object")
        if obj_node is None:
            continue

        obj_name = _get_text(source, obj_node).lower()

        if obj_name in _SAFE_EXEC_OBJECTS:
            continue

        line = prop_node.start_point[0] + 1
        findings.append(
            {
                "rule_id": "SKY-D506",
                "severity": "HIGH",
                "message": "child_process.exec() can lead to command injection. Use execFile() instead.",
                "file": str(file_path),
                "line": line,
                "col": 0,
            }
        )



from __future__ import annotations

from tree_sitter import Language, Parser, Query, QueryCursor
import tree_sitter_typescript as tsts
from skylos.visitor import Definition

try:
    TS_LANG: Language | None = Language(tsts.language_typescript())
except Exception:
    TS_LANG = None

try:
    TSX_LANG: Language | None = Language(tsts.language_tsx())
except Exception:
    TSX_LANG = None

_LIFECYCLE_METHODS: set[str] = {
    "constructor",
    "render",
    "connectedCallback",
    "disconnectedCallback",
    "attributeChangedCallback",
    "componentDidMount",
    "componentWillUnmount",
    "componentDidUpdate",
    "shouldComponentUpdate",
    "getDerivedStateFromProps",
    "getSnapshotBeforeUpdate",
    "ngOnInit",
    "ngOnDestroy",
    "ngOnChanges",
    "ngAfterViewInit",
}


class TypeScriptCore:
    """
    High level wrapper around a tree-sitter TS parse tree that extracts symbol
    definitions and reference occurrences from a single source file.
    """

    def __init__(self, file_path: str, source_bytes: bytes) -> None:
        self.file_path: str = file_path
        self.source: bytes = source_bytes
        self.defs: list[Definition] = []
        self.refs: list[tuple[str, str]] = []
        self.imports: list[dict[str, str | int]] = []

        if str(file_path).endswith(".tsx") and TSX_LANG:
            self.lang: Language | None = TSX_LANG
        else:
            self.lang = TS_LANG

        if self.lang:
            self.parser = Parser(self.lang)
            self.tree = self.parser.parse(source_bytes)
            self.root_node = self.tree.root_node
        else:
            self.tree = None
            self.root_node = None

    def _get_text(self, node) -> str:
        return self.source[node.start_byte : node.end_byte].decode("utf-8")

    def _run_query(self, pattern: str, capture_name: str) -> list:
        if not self.root_node or not self.lang:
            return []

        try:
            query = Query(self.lang, pattern)
            cursor = QueryCursor(query)
            captures = cursor.captures(self.root_node)

            return captures.get(capture_name, [])

        except Exception:
            return []

    _SELF_REF_CONTAINERS: set[str] = {
        "function_declaration",
        "class_declaration",
        "type_alias_declaration",
        "interface_declaration",
        "enum_declaration",
        "variable_declarator",
    }

    def _is_self_ref(self, node, name: str) -> bool:
        current = node.parent
        while current:
            if current.type in self._SELF_REF_CONTAINERS:
                name_node = current.child_by_field_name("name")
                if name_node and self._get_text(name_node) == name:
                    return True
            current = current.parent
        return False

    def _add_ref(self, node) -> None:
        name = self._get_text(node)
        if self._is_self_ref(node, name):
            return
        self.refs.append((name, self.file_path))

    def _add_ref_forced(self, node) -> None:
        self.refs.append((self._get_text(node), self.file_path))

    def scan(self) -> None:
        if not self.root_node:
            self.raw_imports: list[dict] = []
            return

        self._scan_defs()
        self._scan_refs()
        self._scan_imports()
        self._scan_raw_imports()
        self._build_call_graph()

    def _scan_defs(self) -> None:
        for node in self._run_query(
            "(function_declaration name: (identifier) @def)", "def"
        ):
            self._add_def(node, "function")

        for node in self._run_query(
            "(class_declaration name: (type_identifier) @def)", "def"
        ):
            self._add_def(node, "class")

        for node in self._run_query(
            "(interface_declaration name: (type_identifier) @def)", "def"
        ):
            self._add_def(node, "class")

        for node in self._run_query(
            "(enum_declaration name: (identifier) @def)", "def"
        ):
            self._add_def(node, "class")

        for node in self._run_query(
            "(type_alias_declaration name: (type_identifier) @def)", "def"
        ):
            self._add_def(node, "class")

        for node in self._run_query("(decorator (identifier) @dec)", "dec"):
            class_node = node.parent
            if class_node:
                class_node = class_node.parent
            if class_node and class_node.type == "class_declaration":
                name_node = class_node.child_by_field_name("name")
                if name_node:
                    self._add_ref_forced(name_node)

        for node in self._run_query(
            "(decorator (call_expression function: (identifier) @dec))", "dec"
        ):
            decorator_node = node.parent  # call_expression
            if decorator_node:
                decorator_node = decorator_node.parent  # decorator
            if decorator_node:
                class_node = decorator_node.parent  # class_declaration
            else:
                class_node = None
            if class_node and class_node.type == "class_declaration":
                name_node = class_node.child_by_field_name("name")
                if name_node:
                    self._add_ref_forced(name_node)

        for node in self._run_query(
            "(method_definition name: (property_identifier) @def)", "def"
        ):
            self._add_def(node, "method")
        for node in self._run_query(
            "(method_definition name: (identifier) @def)", "def"
        ):
            self._add_def(node, "method")

        for node in self._run_query(
            "(variable_declarator name: (identifier) @def)", "def"
        ):
            var_decl = node.parent  # variable_declarator
            if var_decl:
                value_node = var_decl.child_by_field_name("value")
            else:
                value_node = None
            is_arrow = value_node and value_node.type == "arrow_function"
            if is_arrow:
                self._add_def(node, "function")
            elif self._is_top_level(node):
                self._add_def(node, "variable")

    def _scan_refs(self) -> None:
        call_ref_patterns: list[tuple[str, str]] = [
            ("(call_expression function: (identifier) @ref)", "ref"),
            ("(new_expression constructor: (identifier) @ref)", "ref"),
            ("(member_expression property: (property_identifier) @ref)", "ref"),
        ]

        value_ref_patterns: list[tuple[str, str]] = [
            ("(arguments (identifier) @ref)", "ref"),
            ("(variable_declarator value: (identifier) @ref)", "ref"),
            ("(array (identifier) @ref)", "ref"),
            ("(return_statement (identifier) @ref)", "ref"),
            ("(binary_expression right: (identifier) @ref)", "ref"),
            ("(binary_expression left: (identifier) @ref)", "ref"),
            ("(assignment_expression right: (identifier) @ref)", "ref"),
            ("(spread_element (identifier) @ref)", "ref"),
            ("(member_expression object: (identifier) @ref)", "ref"),
            ("(pair value: (identifier) @ref)", "ref"),
            ("(unary_expression (identifier) @ref)", "ref"),
            ("(template_substitution (identifier) @ref)", "ref"),
        ]

        shorthand_patterns: list[tuple[str, str]] = [
            ("(shorthand_property_identifier) @ref", "ref"),
        ]

        type_ref_patterns: list[tuple[str, str]] = []
        self._scan_type_refs()

        decorator_patterns: list[tuple[str, str]] = [
            ("(decorator (identifier) @ref)", "ref"),
            ("(decorator (call_expression function: (identifier) @ref))", "ref"),
        ]

        export_patterns: list[tuple[str, str]] = [
            ("(export_specifier name: (identifier) @ref)", "ref"),
        ]

        inheritance_patterns: list[tuple[str, str]] = [
            ("(extends_clause (identifier) @ref)", "ref"),
        ]

        all_patterns = (
            call_ref_patterns
            + value_ref_patterns
            + shorthand_patterns
            + type_ref_patterns
            + decorator_patterns
            + export_patterns
            + inheritance_patterns
        )

        for pattern, cap_name in all_patterns:
            for node in self._run_query(pattern, cap_name):
                self._add_ref(node)

    _TYPE_DEF_PARENTS: set[str] = {
        "class_declaration",
        "interface_declaration",
        "enum_declaration",
        "type_alias_declaration",
    }

    def _scan_type_refs(self) -> None:
        for node in self._run_query("(type_identifier) @ref", "ref"):
            parent = node.parent
            if parent and parent.type in self._TYPE_DEF_PARENTS:
                continue
            self._add_ref(node)

    def _find_containing_class(self, node) -> str | None:
        current = node.parent
        while current:
            if current.type == "class_declaration":
                name_node = current.child_by_field_name("name")
                if name_node:
                    return self._get_text(name_node)
            current = current.parent
        return None

    def _add_def(self, node, type_name: str) -> None:
        name = self._get_text(node)

        if type_name == "method" and name in _LIFECYCLE_METHODS:
            return

        if type_name == "method":
            class_name = self._find_containing_class(node)
            if class_name:
                name = f"{class_name}.{name}"

        line = node.start_point[0] + 1

        is_exported = self._is_exported(node)

        d = Definition(name, type_name, self.file_path, line)
        d.is_exported = is_exported
        self.defs.append(d)

    def _is_top_level(self, node) -> bool:
        current = node.parent
        while current:
            if current.type == "program":
                return True
            if current.type in (
                "export_statement",
                "lexical_declaration",
                "variable_declarator",
            ):
                current = current.parent
                continue
            return False
        return False

    def _is_exported(self, node) -> bool:
        try:
            current = node.parent
            for _ in range(4):
                if current is None:
                    break
                if "export" in current.type:
                    return True
                current = current.parent
        except Exception:
            pass
        return False

    def _scan_imports(self) -> None:
        import_patterns = [
            "(import_clause (named_imports (import_specifier name: (identifier) @name)))",
            "(import_clause (identifier) @name)",
            "(import_clause (namespace_import (identifier) @name))",
        ]

        for pattern in import_patterns:
            for node in self._run_query(pattern, "name"):
                name = self._get_text(node)
                line = node.start_point[0] + 1
                d = Definition(name, "import", self.file_path, line)
                self.defs.append(d)
                self.imports.append(
                    {"name": name, "file": str(self.file_path), "line": line}
                )

    def _scan_raw_imports(self) -> None:
        self.raw_imports: list[dict] = []

        for src_node in self._run_query(
            "(import_statement source: (string) @src)", "src"
        ):
            source_path = self._get_text(src_node).strip("'\"")
            import_stmt = src_node.parent
            if import_stmt:
                names = self._extract_import_names_from_stmt(import_stmt)
                self.raw_imports.append(
                    {
                        "source": source_path,
                        "names": names,
                        "line": src_node.start_point[0] + 1,
                    }
                )

        for src_node in self._run_query(
            "(export_statement source: (string) @src)", "src"
        ):
            source_path = self._get_text(src_node).strip("'\"")
            export_stmt = src_node.parent
            if export_stmt:
                names = self._extract_export_names_from_stmt(export_stmt)
                self.raw_imports.append(
                    {
                        "source": source_path,
                        "names": names,
                        "line": src_node.start_point[0] + 1,
                    }
                )

    def _extract_import_names_from_stmt(self, import_stmt) -> list[str]:
        names = []
        for child in import_stmt.children:
            if child.type == "import_clause":
                for clause_child in child.children:
                    if clause_child.type == "named_imports":
                        for spec in clause_child.children:
                            if spec.type == "import_specifier":
                                name_node = spec.child_by_field_name("name")
                                if name_node:
                                    names.append(self._get_text(name_node))
                    elif clause_child.type == "identifier":
                        names.append(self._get_text(clause_child))
                    elif clause_child.type == "namespace_import":
                        names.append("*")
        return names

    def _extract_export_names_from_stmt(self, export_stmt) -> list[str]:
        names = []
        for child in export_stmt.children:
            if child.type == "export_clause":
                for spec in child.children:
                    if spec.type == "export_specifier":
                        name_node = spec.child_by_field_name("name")
                        if name_node:
                            names.append(self._get_text(name_node))
            elif child.type == "*":
                names.append("*")
        return names if names else ["*"]

    def _build_call_graph(self) -> None:
        self.call_pairs: list[tuple[str, str]] = []

        for name_node in self._run_query(
            "(function_declaration name: (identifier) @name)", "name"
        ):
            caller_name = self._get_text(name_node)
            func_node = name_node.parent
            if func_node:
                body = func_node.child_by_field_name("body")
                if body:
                    self._collect_calls_in_body(caller_name, body)

        for name_node in self._run_query(
            "(variable_declarator name: (identifier) @name)", "name"
        ):
            var_decl = name_node.parent
            if var_decl:
                value = var_decl.child_by_field_name("value")
                if value and value.type == "arrow_function":
                    caller_name = self._get_text(name_node)
                    body = value.child_by_field_name("body")
                    if body:
                        self._collect_calls_in_body(caller_name, body)

        for name_node in self._run_query(
            "(method_definition name: (property_identifier) @name)", "name"
        ):
            method_name = self._get_text(name_node)
            class_name = self._find_containing_class(name_node)
            if class_name:
                caller_name = f"{class_name}.{method_name}"
            else:
                caller_name = method_name
            method_node = name_node.parent
            if method_node:
                body = method_node.child_by_field_name("body")
                if body:
                    self._collect_calls_in_body(caller_name, body)

        name_to_def: dict[str, Definition] = {}
        for d in self.defs:
            name_to_def[d.name] = d
            if d.simple_name not in name_to_def:
                name_to_def[d.simple_name] = d

        for caller, callee in self.call_pairs:
            caller_def = name_to_def.get(caller)
            callee_def = name_to_def.get(callee)
            if caller_def and callee_def and caller_def is not callee_def:
                caller_def.calls.add(callee_def.name)
                callee_def.called_by.add(caller_def.name)

    def _collect_calls_in_body(self, caller: str, body_node) -> None:
        stack = [body_node]
        while stack:
            node = stack.pop()
            if node.type == "call_expression":
                func = node.child_by_field_name("function")
                if func:
                    if func.type == "identifier":
                        self.call_pairs.append((caller, self._get_text(func)))
                    elif func.type == "member_expression":
                        prop = func.child_by_field_name("property")
                        obj = func.child_by_field_name("object")
                        if prop:
                            if obj and self._get_text(obj) == "this":
                                class_name = self._find_containing_class(node)
                                if class_name:
                                    self.call_pairs.append(
                                        (caller, f"{class_name}.{self._get_text(prop)}")
                                    )
                            self.call_pairs.append((caller, self._get_text(prop)))
            for child in node.children:
                stack.append(child)



import ast
import json
import tempfile
from pathlib import Path

from skylos.visitor import Visitor
from skylos.implicit_refs import ImplicitRefTracker


class TestGetAttrFStringPattern:
    def test_inline_fstring_registers_pattern(self):
        code = """
import sys

def export_csv(data):
    return ",".join(str(v) for v in data)

def export_json(data):
    import json
    return json.dumps(data)

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("export_" in p for p in patterns)

    def test_inline_fstring_marks_matching_defs(self):
        code = """
import sys

def export_csv(data):
    pass

def export_json(data):
    pass

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name in ("export_csv", "export_json"):
                matched, _, _ = tracker.should_mark_as_used(d)
                assert matched, f"{d.simple_name} should match pattern"

    def test_non_matching_defs_not_marked(self):
        code = """
import sys

def export_csv(data):
    pass

def unrelated_func():
    pass

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name == "unrelated_func":
                matched, _, _ = tracker.should_mark_as_used(d)
                assert not matched, "unrelated_func should not match export_* pattern"


class TestGlobalsFStringPattern:
    def test_globals_fstring_registers_pattern(self):
        code = """
def handle_create(payload):
    return {"action": "created"}

def handle_delete(payload):
    return {"action": "deleted"}

HANDLER_MAP = {a: globals()[f"handle_{a}"] for a in ("create", "delete")}
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("handle_" in p for p in patterns)

    def test_globals_fstring_marks_matching_defs(self):
        code = """
def handle_create(payload):
    pass

def handle_update(payload):
    pass

HANDLER_MAP = {a: globals()[f"handle_{a}"] for a in ("create", "update")}
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name in ("handle_create", "handle_update"):
                matched, _, _ = tracker.should_mark_as_used(d)
                assert matched, f"{d.simple_name} should match pattern"


class TestInitSubclassRegistry:
    def test_base_classes_populated(self):
        code = """
class Base:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

class Child(Base):
    pass
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        child = None
        for d in v.defs:
            if d.simple_name == "Child":
                child = d
                break

        assert child is not None
        assert "mymod.Base" in child.base_classes

    def test_subclasses_not_flagged_unused(self):
        code = """
_REGISTRY = {}

class RegisteredHandler:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _REGISTRY[cls.name] = cls

class EmailHandler(RegisteredHandler):
    name = "email"

class SlackHandler(RegisteredHandler):
    name = "slack"

def get_handler(name):
    return _REGISTRY.get(name)
"""
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "registry.py"
            p.write_text(code)

            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0))
            unused_classes = {
                item["simple_name"] for item in result.get("unused_classes", [])
            }

            assert "EmailHandler" not in unused_classes
            assert "SlackHandler" not in unused_classes


class TestEnumMemberDetection:
    def _get_unused(self, code):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "mod.py"
            p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0))
            unused = set()
            for cat in ["unused_functions", "unused_classes", "unused_variables"]:
                for item in result.get(cat, []):
                    unused.add(item["simple_name"])
            return unused

    def test_basic_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import Enum

class Color(Enum):
    RED = 1
    GREEN = 2
    BLUE = 3

c = Color.RED
""")
        assert "RED" not in unused
        assert "GREEN" not in unused
        assert "BLUE" not in unused

    def test_int_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import IntEnum

class Status(IntEnum):
    PENDING = 0
    ACTIVE = 1

s = Status.ACTIVE
""")
        assert "PENDING" not in unused
        assert "ACTIVE" not in unused

    def test_enum_methods_not_flagged(self):
        unused = self._get_unused("""
from enum import Enum

class Color(Enum):
    RED = 1
    GREEN = 2

    def hex_code(self):
        return f"#{self.value:06x}"

c = Color.RED
""")
        assert "hex_code" not in unused

    def test_indirect_enum_inheritance(self):
        unused = self._get_unused("""
from enum import Enum

class BaseEnum(Enum):
    pass

class Priority(BaseEnum):
    LOW = 1
    HIGH = 2

p = Priority.HIGH
""")
        assert "LOW" not in unused
        assert "HIGH" not in unused

    def test_flag_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import Flag

class Permission(Flag):
    READ = 1
    WRITE = 2
    EXECUTE = 4

p = Permission.READ | Permission.WRITE
""")
        assert "READ" not in unused
        assert "WRITE" not in unused
        assert "EXECUTE" not in unused
